#!/usr/bin/env nextflow

params.fasta = ""
params.working = ""
params.databases = ""
params.threads = 1
params.evalue = 10
params.splitby = 100000
params.bitscore_cutoff = false
params.log = '/dev/null'
params.help = false
params.slurm = false
params.ancestral = false
params.includecc = false 
ancestral_flag = ""
include_cc_flag = ""
params.prefix = ''
cutoff_flag = "--cutoff=5"
mode="ont"

Executor = 'local'

MODULES =  "$workflow.projectDir/../modules"
SCRIPTS =  "$workflow.projectDir/../scripts"


def usage() {
    log.info ""
    log.info "Usage: nextflow run nanopore.nf --fasta /Path/to/infile.fasta --databases /Path/to/databases --working /Path/to/working_directory [--evalue 10] [--threads 4] [--slurm]  [--log /Path/to/log.txt]"
    log.info "  --fasta     Path to the input NT FASTA file"
    log.info "  --databases Path to the database directory"
    log.info "  --working   Path to the output working directory"
    log.info "  --evalue    E value cut-off (Default=10)"
    log.info "  --threads   Number of threads to use (Default=1)"
    log.info "  --splitby   Number of fasta sequences in each chunk"
    log.info "  --slurm     Submit modules in this workflow to run on a SLURM grid (Default = run locally)"
    log.info "  --ancestral Include all ancestral GO terms" 
    log.info "  --includecc     Include cellular component go terms"
    log.info "  --bitscore_cutoff Tiebreak across all uniprots within this % of the top bitscore"
    log.info "  --prefix   Add argument to beginning of the output files"
    log.info "  --version   Verison number for the seqscreen pipeline"
    log.info "  --log       Where to write log file (Default=no log)"
    log.info "  -h, --help  Print this help message out"
    log.info ""
    exit 1
}


if (params.help) {
    usage()
}

if (params.slurm) {
    Executor = 'slurm'
}

if (!params.fasta) {
    log.error "Missing argument for --fasta option"
    usage()
}

if (!params.working) {
    log.error "Missing argument for --working option"
    usage()
}

if (!params.databases) {
    log.error "Missing argument for --databases option"
    usage()
}
rflag = 'ont'
if (params.bitscore_cutoff) {
    cutoff_flag = "--cutoff=${params.bitscore_cutoff}"
    if(rflag) {
        rflag = rflag + ",bitscore:" + params.bitscore_cutoff
    }else{
        rflag = "bitscore:" + params.bitscore_bitscore_cutoff
    }

}

if (params.evalue != "") {
   if(rflag) {
       rflag = rflag + ",evalue:" + params.evalue
   }
   else {
       rflag = "evalue:" + params.evalue
   }
}


if (params.includecc) {
    include_cc_flag = "--include-cc"
    if(rflag){
        rflag = rflag + ",includecc"
    }else{
        rflag = 'includecc'
    }

}

if (params.ancestral) {
    ancestral_flag = "--ancestral"
    if(rflag) {
      
       rflag = rflag + ",ancestral"
    }
   else {
       rflag = "ancestral"
   }
}

if (params.splitby != ""){
    if(rflag){
        rflag = rflag + ",splitby:" + params.splitby
    }else{
        rflag = "splitby:" + params.splitby
    }
}



if(! rflag)
{
   rflag = "None"
}



fastafile = file(params.fasta)
workingDir = file(params.working)
databaseDir = file(params.databases)
versionNum = params.version
logfile = file(params.log)

BASE = fastafile.getName()
THREADS = params.threads
EVALUE = params.evalue
WORKFLOW = "taxonomic_identification"
FWORKFLOW="functional_annotation"
RWORKFLOW="report_generation"

blastnDir = file("$workingDir/$WORKFLOW/blastn")
blastxDir = file("$workingDir/$WORKFLOW/blastx")
taxDir = file("$workingDir/$WORKFLOW/taxonomic_assignment")
fxnDir = file("$workingDir/$FWORKFLOW/functional_assignments")

process Create_Working_Directories {
    output:
    stdout create

    """
    mkdir -p $workingDir/$WORKFLOW/
    mkdir -p $workingDir/$FWORKFLOW
    mkdir -p $workingDir/$RWORKFLOW
    if [ -d $blastnDir ]; then rm -rf $blastnDir; fi;
    if [ -d $blastxDir ]; then rm -rf $blastxDir; fi;
    if [ -d $taxDir ]; then rm -rf $taxDir; fi;    
    if [ -d $fxnDir ]; then rm -rf $fxnDir; fi;    
    
    mkdir $blastnDir
    mkdir $blastxDir
    mkdir $taxDir
    mkdir $fxnDir
    """
}

process Initialize {
    input:
    val create from create

    output:
    stdout init

    executor Executor

    """
    echo -n " # Launching $WORKFLOW workflow ...... " | tee -a $logfile; date '+%H:%M:%S %Y-%m-%d' | tee -a $logfile
    """
}

Channel
    .fromPath(fastafile)
    .splitFasta(by: params.splitby, file: true)
    .set { fasta_ch }

process mmseqs2 {

    input:
    val	init from init
    file 'chunk' from fasta_ch

    output:
    file 'outchunk.ur100.btab' into mmseqs2BtabResults
    stdout mmseqs2_std into mmseqs2_wait

    if ( Executor == 'local' ) {
       executor "local"
       maxForks 1
    }

    else if ( Executor == 'slurm' ) {
       clusterOptions "--ntasks-per-node $THREADS"
       executor "slurm"
    }

    script:
    """
    echo "running MMSeqs2 on ${chunk}"
 
    mmseqs createdb --dbtype 2 --shuffle 0 chunk "${blastxDir}/chunkDB"

    echo "Created sequence DB"

    rm -rf $blastxDir/searchDB.*

    mmseqs search "${blastxDir}/chunkDB" \
    "${databaseDir}/mmseqs2/uniref.mini.mmseqs2" \
    "${blastxDir}/searchDB" \
    "${blastxDir}/tmp" \
    --threads "${THREADS}" \
    -s 4 \
    --headers-split-mode 1 \
    --min-aln-len 100 \
    -e "${EVALUE}" \
    --split 0 \
    --split-memory-limit 20G \
    --split-mode 0
    
    mmseqs convertalis "${blastxDir}/chunkDB" \
    "${databaseDir}/mmseqs2/uniref.mini.mmseqs2" \
    "${blastxDir}/searchDB" \
    outchunk.ur100.btab \
    --threads "${THREADS}" \
    --format-output "query,target,pident,alnlen,mismatch,gapopen,qstart,qend,tstart,tend,evalue,bits,fident,qframe,raw,theader"
   
    echo "Finish converting format"
    
    
    """
}
mmseqs2BtabResults.into{ postprocessMmseqs2_trigger; mmseqs2BtabResults_coll}
mmseqs2BtabResults_coll.collectFile(name: "${blastxDir}/${BASE}.ur100.btab").set { mmseqs2_collect_ch }

process postprocessMmseqs2 {
    input: 
    file "*outchunk.ur100.btab" from postprocessMmseqs2_trigger.collect()

    output:
    stdout postprocessMmseqs2_out

      if (Executor == 'local') {                                        
          executor "local"                                              
      }                                                                 
                                                                        
      else if (Executor == "slurm") {                                   
          clusterOptions "--ntasks-per-node $THREADS"                   
          executor "slurm"                                              
      }                                                                 
      script: 
      """
        ln -s ${BASE}.ur100.btab ${blastxDir}/functional_link.ur100.btab
        touch ${blastxDir}/functional_link.ur100.xml
      """
}

process splitReads {
    input:
    val postprocessMmseqs2_out

    output:
    stdout splitreads

    executor Executor

    """
      python3 ${SCRIPTS}/split_reads.py --fasta=$fastafile --blastx=${blastxDir}/${BASE}.ur100.btab
    """

}

process taxonomic_assignment {
    input:
    val pp from postprocessMmseqs2_out
    val sr from splitreads

    output:
    stdout taxOut

    executor Executor
    
    """
    ${SCRIPTS}/btab_2_tax_report_sensitive.pl --blastx=${blastxDir}/${BASE}.ur100.split.btab \
                                              --blastn=${blastnDir}/${BASE}.nt.btab_outlier_clean.btab \
    				                          --out=${taxDir}/taxonomic_results.txt \
    				                          $cutoff_flag
    echo -n " # $WORKFLOW workflow complete ....... " | tee -a $logfile; date '+%H:%M:%S %Y-%m-%d' | tee -a $logfile
    """
}

process functional_assignments {
    input:
    val init from init
    val pp from postprocessMmseqs2_out
    val sr from splitreads

    output:
    stdout functional_assignments

    executor Executor
    
    """
    echo -n " # Launching $FWORKFLOW workflow ...... " | tee -a $logfile; date '+%H:%M:%S %Y-%m-%d' | tee -a $logfile
    while [ ! -f $workingDir/taxonomic_identification/blastx/${BASE}.ur100.split.btab ]
    do
      sleep 5
    done

    ${SCRIPTS}/functional_report_generation.py --fasta=$workingDir/taxonomic_identification/blastx/${BASE}.split.fasta \
                                               --urbtab=$workingDir/taxonomic_identification/blastx/${BASE}.ur100.split.btab \
                                               --go=${databaseDir}/go/go_network.txt \
                                               --out=${workingDir}/$FWORKFLOW/functional_assignments/functional_results.txt \
                                               --annotation=${databaseDir}/annotation_scores.pck \
                                               $ancestral_flag \
                                               $include_cc_flag \
                                               $cutoff_flag > ${workingDir}/$FWORKFLOW/functional_assignments.log 2>&1
    echo -n " # $FWORKFLOW workflow complete ....... " | tee -a $logfile; date '+%H:%M:%S %Y-%m-%d' | tee -a $logfile
    """
}

process seqscreen_tsv_report {
    input:
	val init from init
    val tx from taxOut
    val fa from functional_assignments
    
    output:
	stdout seqscreen_tsv_report

    executor Executor
    
    """
    echo -n " # Launching $RWORKFLOW workflow ...... " | tee -a $logfile; date '+%H:%M:%S %Y-%m-%d' | tee -a $logfile
    ${SCRIPTS}/seqscreen_tsv_report.py \
                                    --taxonomy=${workingDir}/taxonomic_identification/taxonomic_assignment/taxonomic_results.txt \
                                    --functional=${workingDir}/functional_annotation/functional_assignments/functional_results.txt \
                                    --taxlookup=$databaseDir/taxonomy/taxa_lookup.txt \
                                    --funsocs=$databaseDir/funsocs \
                                    --fasta=$workingDir/taxonomic_identification/blastx/${BASE}.split.fasta \
                                    --mode=$mode \
                                    --out=$workingDir/$RWORKFLOW/${params.prefix}seqscreen_report.tsv
    """
}

process seqscreen_html_report {
    input:
        val seqscreen_tsv_report from seqscreen_tsv_report

    output:
        stdout seqscreen_html_report

    executor Executor

    """
    ${MODULES}/html_report_generation.sh \
        --report=$workingDir/$RWORKFLOW/${params.prefix}seqscreen_report.tsv \
        --fasta=$workingDir/taxonomic_identification/blastx/${BASE}.split.fasta \
        --blastx=$workingDir/taxonomic_identification/blastx/functional_link.ur100.xml \
        --version=$versionNum \
        --mode=$mode \
        --rflag=$rflag \
        --gonetwork=$databaseDir/go/go_network.txt \
        --out=$workingDir/$RWORKFLOW/${params.prefix}seqscreen_html_report/
    echo -n " # $RWORKFLOW workflow complete ....... " | tee -a $logfile; date '+%H:%M:%S %Y-%m-%d' | tee -a $logfile
    """
}

process cleanup {
    input:
    val seqscreen_html_report
    
    output:
    stdout cleanup

    script:
        """
        echo -n " # Launching cleanup workflow ...... " | tee -a $logfile; date '+%H:%M:%S %Y-%m-%d' | tee -a $logfile
        rm -r $workingDir/taxonomic_identification/blastn
        rm -rf $blastxDir/searchDB.*
        rm -rf $blastxDir/chunkDB*
        echo -n " # cleanup complete ....... " | tee -a $logfile; date '+%H:%M:%S %Y-%m-%d' | tee -a $logfile
        """
}



create.subscribe { print "$it" }
init.subscribe { print "$it" }
postprocessMmseqs2_out.subscribe { print "$it" }
splitreads.subscribe { print "$it" }
taxOut.subscribe { print "$it" }
functional_assignments.subscribe { print "$it" }
seqscreen_tsv_report.subscribe { print "$it" }
seqscreen_html_report.subscribe { print "$it" }
cleanup.subscribe { print "$it" }
